# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import datetime
import logging
import os.path

from bs_lib.Pod   import Pod

from bs_loco.xloco import xLoco

from bs_loco.db.tables      import tMsg

from .provider_sms import ProviderSms


class LogfileSms(ProviderSms):
    """
    This is the logfile sms provider.
    """
    LOG_FORMAT       = '[%(asctime)s.%(msecs)03d] [%(levelname)s] %(message)s'
    CONSOLE_FORMAT   = '%(message)s '
    TIMESTAMP_FORMAT = '%Y-%m-%d %H:%M:%S'

    def __init__(self):
        """
        Constructor.
        """
        ProviderSms.__init__(self)
        self._logfile = None


    def id(self) -> str:
        """
        Overload.
        """
        return 'logfile'


    def initialize(self, pod: Pod, dao: object, cfg: dict):
        """
        Overload.
        """
        ProviderSms.initialize(self, pod, dao, cfg)

        self._logfile = self._init_logger(cfg)


    def destroy(self):
        """
        Overload.
        """
        self._logfile = None
        ProviderSms.destroy(self)


    def _provider_send(self, payload: dict, res_list: list, msg: tMsg, addr_list: list) -> tuple:
        """
        Overload.
        """
        try:
            dt_start = datetime.datetime.now()

            self._logfile.info('Sending SMS [msg:%d, to:%s]: %s' % (msg.id, str(addr_list), payload['msg']))

            dt_end = datetime.datetime.now()
        except Exception as x:
            self._log.error('Exception caught sending sms [%s]' % (str(msg)))
            self._log.exception(x)
            return (False, None, str(x))

        return (True, dt_end - dt_start, None)


    def _init_logger(self, cfg: dict):
        """
        Initialize the logger object.

        :param cfg: The config.
        """
        if not cfg.get('path'):
            raise xLoco('Invalid config, [path] not set [provider:%s, cfg:%s]' % (self.name(), str(cfg)))

        log_level = cfg.get('log_level') or 'info'
        filename  = os.path.expandvars(cfg['path'])

        dirpath = os.path.split(filename)

        if dirpath[0] and not os.path.exists(dirpath[0]):
            raise xLoco('Invalid config, directory not found for [path] [provider:%s, path:%s]' % (self.name(), str(filename)))

        logging.basicConfig(level=getattr(logging, log_level.upper()))

        logger           = logging.getLogger(self.name())
        logger.propagate = False

        from logging.handlers import RotatingFileHandler

        file_handler = RotatingFileHandler(filename, 'a', 1 * 1024 * 1024, 10)
        file_handler.setFormatter(logging.Formatter(self.LOG_FORMAT, self.TIMESTAMP_FORMAT))
        file_handler.setLevel(getattr(logging, log_level.upper()))
        logger.addHandler(file_handler)

        if cfg.get('stdout'):
            import sys

            stream_handler = logging.StreamHandler(sys.stdout)
            stream_handler.setFormatter(logging.Formatter(self.CONSOLE_FORMAT, self.TIMESTAMP_FORMAT))
            stream_handler.setLevel(getattr(logging, log_level.upper()))
            logger.addHandler(stream_handler)

        return logger



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.corr_type import tCorrType

class dCorrTypeSelectAll:

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.orec     = tCorrType()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    async def fetch(self) -> bool:
        """
        Fetches the next row from the executed query.

        :return: True if a row is retreived.
        """
        if not await self._dbcon.fetch(self._dbstmnt):
            return False

        self.orec.id = self._dbstmnt.result[0]
        self.orec.descr = self._dbstmnt.result[1]
        self.orec.status = self._dbstmnt.result[2]
        self.orec.templ_schema = self._dbstmnt.result[3]
        self.orec.modified_by = self._dbstmnt.result[4]
        self.orec.tm_stamp = self._dbstmnt.result[5]

        return True

    async def fetch_all(self, outList: tCorrType.List, clearList: bool = True) -> int:
        """
        Fetch all the rows into the specified list and return the list count.

        :param outList:
        :param clearList:
        :return: Number of rows fetched.
        """
        if clearList:
            outList.clear()

        while await self.fetch():
            outList.append(copy.deepcopy(self.orec))

        return len(outList)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.imsg_req_busy_changed import iMsgReqBusyChanged
from bs_loco.db.tables.omsg_req_busy_changed import oMsgReqBusyChanged

class dMsgReqBusyChanged:

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = iMsgReqBusyChanged()
        self.orec     = oMsgReqBusyChanged()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    def exec_deft(self,
                  limit_crit: str) -> "dMsgReqBusyChanged":
        """
        Execute the query by setting all the inputs.

        :param limit_crit: str
        :return: Self for convenience.
        """
        self.irec.limit_crit = limit_crit

        return self.exec()

    def exec(self, irec: iMsgReqBusyChanged = None) -> "dMsgReqBusyChanged":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("MsgReqBusyChanged", self._dbcon.STMNT_TYPE_READ)

        self._dbstmnt.sql("""select
  r.id,
  m.status,
  max(m.time_sent) as timesent
from
  (
    select xr.id
    from loco.msgreq xr join loco.msg xm on xr.id = xm.msgreq_id
    where xr.status = 'b' and xm.status not in ('P', 'R')
    group by xr.id
      except
    select xr.id
    from loco.msgreq xr join loco.msg xm on xr.id = xm.msgreq_id
    where xr.status = 'b' and xm.status in ('P', 'R')
    group by xr.id
  ) r
    join
  loco.msg m
    on m.msgreq_id = r.id
group by
  r.id,
  m.status
order by
  1,
  3
[limitCrit]""")

        self._dbstmnt.dynamic("[limitCrit]", self.irec.limit_crit)

        self._dbstmnt.bind_out("msgReqId", int, 8)
        self._dbstmnt.bind_out("msgStatus", str, 1)
        self._dbstmnt.bind_out("msgTime", datetime.datetime)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self

    def fetch(self) -> bool:
        """
        Fetches the next row from the executed query.

        :return: True if a row is retreived.
        """
        if not self._dbcon.fetch(self._dbstmnt):
            return False

        self.orec.msg_req_id = self._dbstmnt.result[0]
        self.orec.msg_status = self._dbstmnt.result[1]
        self.orec.msg_time = self._dbstmnt.result[2]

        return True

    def fetch_all(self, outList: oMsgReqBusyChanged.List, clearList: bool = True) -> int:
        """
        Fetch all the rows into the specified list and return the list count.

        :param outList:
        :param clearList:
        :return: Number of rows fetched.
        """
        if clearList:
            outList.clear()

        while self.fetch():
            outList.append(copy.deepcopy(self.orec))

        return len(outList)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#

from .config_key import tConfigKey
from .config import tConfig
from .iconfig_search import iConfigSearch
from .corr_prov_key import tCorrProvKey
from .corr_prov import tCorrProv
from .icorr_prov_for_service import iCorrProvForService
from .ocorr_prov_for_service import oCorrProvForService
from .corr_prov_cfg_key import tCorrProvCfgKey
from .corr_prov_cfg import tCorrProvCfg
from .corr_type_key import tCorrTypeKey
from .corr_type import tCorrType
from .corr_type_allow_key import tCorrTypeAllowKey
from .corr_type_allow import tCorrTypeAllow
from .icorr_type_allow_regex_for_type import iCorrTypeAllowRegexForType
from .ocorr_type_allow_regex_for_type import oCorrTypeAllowRegexForType
from .msg_key import tMsgKey
from .msg import tMsg
from .imsg_by_msg_req import iMsgByMsgReq
from .imsg_for_processing import iMsgForProcessing
from .omsg_for_processing import oMsgForProcessing
from .imsg_status_cnt_by_req import iMsgStatusCntByReq
from .omsg_status_cnt_by_req import oMsgStatusCntByReq
from .msg_hist_key import tMsgHistKey
from .msg_hist import tMsgHist
from .imsg_hist_search import iMsgHistSearch
from .msg_req_key import tMsgReqKey
from .msg_req import tMsgReq
from .imsg_req_update_status import iMsgReqUpdateStatus
from .imsg_req_pending_changed import iMsgReqPendingChanged
from .omsg_req_pending_changed import oMsgReqPendingChanged
from .imsg_req_busy_changed import iMsgReqBusyChanged
from .omsg_req_busy_changed import oMsgReqBusyChanged
from .imsg_req_for_archive import iMsgReqForArchive
from .msg_req_res_key import tMsgReqResKey
from .msg_req_res import tMsgReqRes
from .imsg_req_res_by_msg_req import iMsgReqResByMsgReq
from .no_dict_key import tNoDictKey
from .no_dict import tNoDict
from .ino_dict_search import iNoDictSearch
from .ono_dict_search import oNoDictSearch
from .no_res_key import tNoResKey
from .no_res import tNoRes
from .ino_res_search import iNoResSearch
from .ono_res_search import oNoResSearch
from .no_templ_key import tNoTemplKey
from .no_templ import tNoTempl
from .ino_templ_by_no_type import iNoTemplByNoType
from .no_templ_cfg_key import tNoTemplCfgKey
from .no_templ_cfg import tNoTemplCfg
from .no_type_key import tNoTypeKey
from .no_type import tNoType
from .snap_msg_req_key import tSnapMsgReqKey
from .snap_msg_req import tSnapMsgReq
from .snap_prov_key import tSnapProvKey
from .snap_prov import tSnapProv



# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #


def dao_by_name(dao_name: str, dao_path: str = 'dao'):
    """
    Loads a dao object by name:

    :param dao_name: The dao library to load.
    :param dao_path: Optionally provide an alternate dao path
    :return: The dao library.
    """
    from bs_lib import common

    mpath = 'bs_loco.db.%s.%s' % (dao_path, dao_name)

    res, exc = common.import_dyn_pluggin(mpath, True, False)

    if exc:
        raise Exception('DAO not found [%s] - Error: %s' % (dao_name, str(exc)))

    return res


def get_table_sql(db_name: str) -> str:
    """
    Gets the creat tables sql for the targetted database.

    :param db_name: The database to get for, note the equivalent sql file should exist in the package sql folder.
    :returns: The sql file text.
    """
    import importlib.resources as pkg_resources

    from . import sql

    return pkg_resources.read_text(sql, f'{db_name.lower()}.sql')


def _cli_show_usage():
    print()
    print('usage: bs-loco [--getsql] DBNAME')
    print()
    print('required arguments:')
    print('  DBNAME          The target database to get for [eg: postgresql]')
    print()
    print('optional arguments:')
    print('  -g, --getsql    Gets the sql to generate the loco tables')
    print('  -?, -h, --help  Show this help message and exit')
    print()


def _cli_read_args(args: list) -> dict:
    import getopt

    try:
        optlist, gargs = getopt.getopt(args, '?hg', [ 'help', 'getsql' ])
    except getopt.GetoptError as err:
        print(err)
        return _cli_show_usage()

    aobj = { 'db'  : None, 'cmd' : None }

    for o, a in optlist:
        if o in ('-?', '-h', '--help'):
            return _cli_show_usage()

        if o in ('-g', '--getsql'):
            aobj['cmd'] = 'get'
            continue

        print('\n ... unrecognized arguments: %s' % o)
        return _cli_show_usage()

    if not gargs:
        print('\n ... target database not specified')
        return _cli_show_usage()

    if not aobj['cmd']:
        print('\n ... command not specified')
        return _cli_show_usage()


    aobj['db'] = gargs[0]

    return aobj


def main(args: list = None) -> int:
    import sys

    aobj = _cli_read_args(sys.argv[1:])

    if not aobj:
        return 2

    if aobj['cmd'] == 'get':
        print()
        print(get_table_sql(aobj['db']))
        print()

    return 0


__all__ = [ 'dao_by_name', 'get_table_sql' ]



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import asyncio
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.msg_req_res import tMsgReqRes
from bs_loco.db.tables.msg_req_res_key import tMsgReqResKey

from .msg_req_res_lock_one import dMsgReqResLockOne
from .msg_req_res_select_one import dMsgReqResSelectOne
from .msg_req_res_update import dMsgReqResUpdate
from .msg_req_res_insert import dMsgReqResInsert
from .msg_req_res_delete_one import dMsgReqResDeleteOne

class dMsgReqRes:

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self.rec    = tMsgReqRes()
        self._dbcon = dbcon

    def __enter__(self):
        """
        DAO enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        DAO exit.
        """
        pass

    async def lock_one(self, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Lock and select the record.

        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        _key = tMsgReqResKey()

        self.rec._write_key(_key)

        return await self.lock_one_by_key(_key, _lock4update, _expected)

    async def lock_one_with(self, _rec: tMsgReqRes, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Lock and select the record using a record.

        :param _rec: The table rec to select with and into.
        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        _key = tMsgReqResKey()
        _rec._write_key(_key)

        await self.lock_one_by_key(_key, _lock4update, _expected)
        _rec._copy_from(self.rec)
        return True

    async def lock_one_deft(self,
                            id: int,
                            _lock4update: mettle.db.DBLock,
                            _expected: bool = True) -> bool:
        """
        Attempt to lock and select the record by it's primary key columns.

        :param id: int
        :param _lock4update:  The db lock object to use. 
        :param _expected: bool
        """
        _key = tMsgReqResKey()

        _key.id = id

        return await self.lock_one_by_key(_key, _lock4update, _expected)

    async def lock_one_by_key(self, _key: tMsgReqResKey, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Attempt to lock and select the record with it's primary key.

        :param _key: Primary key of the table record.
        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        with dMsgReqResLockOne(self._dbcon) as _qry:
            _dex = _lock4update.retrys - 1 if _lock4update.retrys > 0 else 0

            while _dex >= 0:
                try:
                    await _qry.exec(_key)
                    if not await _qry.fetch():
                        if not _expected:
                            return False

                        raise mettle.lib.xMettle("LockOne failed to return a record (dMsgReqRes)")

                    break
                except mettle.lib.xMettle as _x:
                    if _x.get_error_code() != mettle.lib.xMettle.eCode.DBLockNoWaitFailed:
                        raise _x

                    if _dex == 0:
                        raise mettle.lib.xMettle("Could not lock MsgReqRes record after %d attempt(s) at %d mili second intervals!" % (
_lock4update.retrys, _lock4update.mili_seconds), errCode = mettle.lib.xMettle.eCode.DBLockNoWaitFailed)

                    if _lock4update.mili_seconds > 0:
                        await asyncio.sleep(_lock4update.mili_seconds * 0.001)
                _dex -= 1

            self.rec = _qry.orec

            return True

    async def try_select_one(self) -> bool:
        """
        Attempt to select the record.

        :return: True if the record was found.
        """
        _key = tMsgReqResKey()

        self.rec._write_key(_key)

        return await self.try_select_one_by_key(_key)

    async def try_select_one_with(self, _rec: tMsgReqRes) -> bool:
        """
        Attempt to select the record using a table rec.

        :param _rec: The table rec to select with and into.
        :return: True if the row as selectd.
        """
        _key = tMsgReqResKey()
        _rec._write_key(_key)

        if not await self.try_select_one_by_key(_key):
            return False

        _rec._copy_from(self.rec)

        return True

    async def try_select_one_deft(self,
                                  id: int) -> bool:
        """
        Attempt to select the record by it's primary key columns.

        :param id: int
        :return: True if the row as selectd.
        """
        _key = tMsgReqResKey()

        _key.id = id

        return await self.try_select_one_by_key(_key)

    async def try_select_one_by_key(self, _key: tMsgReqResKey) -> bool:
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :returns: True if row was selected.
        """
        with dMsgReqResSelectOne(self._dbcon) as _qry:
            await _qry.exec(_key)
            if not await _qry.fetch():
                return False

            self.rec = _qry.orec

        return True

    async def select_one(self) -> "dMsgReqRes":
        """
        Selects the record, raises and error if it does not exist.

        :return: Self for convenience.
        """
        if not await self.try_select_one():
            raise mettle.lib.xMettle("SelectOne failed to return a record (dMsgReqRes)")

        return self

    async def select_one_with(self, _rec: tMsgReqRes) -> "dMsgReqRes":
        """
        Attempt to select the record with the table record.

        :param _rec: Table record.
        :returns: Self for convenience.
        """
        if not await self.try_select_one_with(_rec):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dMsgReqRes)")

        return self

    async def select_one_deft(self,
                              id: int) -> "dMsgReqRes":
        """
        Selects the record by its primary key columns, raises and error if it does not exist.

        :param id: int
        """
        _key = tMsgReqResKey()

        _key.id = id

        if not await self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dMsgReqRes)")

        return self

    async def select_one_by_key(self, _key: tMsgReqResKey) -> "dMsgReqRes":
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :return: Self for convenience.
        """
        if not await self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dMsgReqRes)")

        return self

    async def update_deft(self,
                          id: int,
                          msgreq_id: int,
                          corr_id_list: dict,
                          filename: str,
                          mimetype: str,
                          content: bytearray,
                          modified_by: str):
        """
        Update the record by setting all column values.

        :param id: int
        :param msgreq_id: int
        :param corr_id_list: dict
        :param filename: str
        :param mimetype: str
        :param content: bytearray
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        """
        self.rec.id = id
        self.rec.msgreq_id = msgreq_id
        self.rec.corr_id_list = corr_id_list
        self.rec.filename = filename
        self.rec.mimetype = mimetype
        self.rec.content = content
        self.rec.modified_by = modified_by

        await self.update()

    async def update(self, irec: tMsgReqRes = None):
        """
        Update the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to update the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dMsgReqResUpdate(self._dbcon) as _qry:
            await _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)

    async def insert_deft(self,
                          msgreq_id: int,
                          corr_id_list: dict,
                          filename: str,
                          mimetype: str,
                          content: bytearray,
                          modified_by: str):
        """
        Insert the record setting all column values.

        :param id: int
        :param msgreq_id: int
        :param corr_id_list: dict
        :param filename: str
        :param mimetype: str
        :param content: bytearray
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        """
        self.rec.msgreq_id = msgreq_id
        self.rec.corr_id_list = corr_id_list
        self.rec.filename = filename
        self.rec.mimetype = mimetype
        self.rec.content = content
        self.rec.modified_by = modified_by

        await self.insert()

    async def insert(self, irec: tMsgReqRes = None):
        """
        Insert the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to insert into the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dMsgReqResInsert(self._dbcon) as _qry:
            await _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)

    async def delete_one_deft(self,
                              id: int):
        """
        Delete one record by it's primary key columns.

        :param id: int
        """
        _key = tMsgReqResKey()

        _key.id = id

        self.delete_one(_key)

    async def delete_one(self, irec: tMsgReqResKey = None):
        """
        Delete the record by its primary key or irec if it is not None.

        :param irec: If not None use this record to delete record.
        """
        _key = tMsgReqResKey()

        if irec:
            _key._copy_from(irec)
        else:
            self.rec._write_key(_key)

        with dMsgReqResDeleteOne(self._dbcon) as _qry:
            await _qry.exec(_key)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.snap_msg_req import tSnapMsgReq

class dSnapMsgReqUpdate:

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tSnapMsgReq()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    def exec_deft(self,
                  valuedate: datetime.date,
                  notype_id: str,
                  tot_req: int,
                  tot_proc: int,
                  tot_cancelled: int,
                  tot_part_canc: int,
                  tot_msgs: int,
                  tot_msgs_sent: int,
                  tot_msgs_failed: int,
                  tot_msgs_retrys: int) -> "dSnapMsgReqUpdate":
        """
        Execute the query by setting all the inputs.

        :param valuedate: datetime.date
        :param notype_id: str
        :param tot_req: int
        :param tot_proc: int
        :param tot_cancelled: int
        :param tot_part_canc: int
        :param tot_msgs: int
        :param tot_msgs_sent: int
        :param tot_msgs_failed: int
        :param tot_msgs_retrys: int
        :return: Self for convenience.
        """
        self.irec.valuedate = valuedate
        self.irec.notype_id = notype_id
        self.irec.tot_req = tot_req
        self.irec.tot_proc = tot_proc
        self.irec.tot_cancelled = tot_cancelled
        self.irec.tot_part_canc = tot_part_canc
        self.irec.tot_msgs = tot_msgs
        self.irec.tot_msgs_sent = tot_msgs_sent
        self.irec.tot_msgs_failed = tot_msgs_failed
        self.irec.tot_msgs_retrys = tot_msgs_retrys

        return self.exec()

    def exec(self, irec: tSnapMsgReq = None) -> "dSnapMsgReqUpdate":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("SnapMsgReqUpdate", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""update
  loco.SnapMsgReq
set
  tot_req = :tot_req,
  tot_proc = :tot_proc,
  tot_cancelled = :tot_cancelled,
  tot_part_canc = :tot_part_canc,
  tot_msgs = :tot_msgs,
  tot_msgs_sent = :tot_msgs_sent,
  tot_msgs_failed = :tot_msgs_failed,
  tot_msgs_retrys = :tot_msgs_retrys
where
  valuedate = :valuedate and
  notype_id = :notype_id""")

        self._dbstmnt.bind_in("valuedate", self.irec.valuedate, datetime.date)
        self._dbstmnt.bind_in("notype_id", self.irec.notype_id, str, 64)
        self._dbstmnt.bind_in("tot_req", self.irec.tot_req, int, 4)
        self._dbstmnt.bind_in("tot_proc", self.irec.tot_proc, int, 4)
        self._dbstmnt.bind_in("tot_cancelled", self.irec.tot_cancelled, int, 4)
        self._dbstmnt.bind_in("tot_part_canc", self.irec.tot_part_canc, int, 4)
        self._dbstmnt.bind_in("tot_msgs", self.irec.tot_msgs, int, 4)
        self._dbstmnt.bind_in("tot_msgs_sent", self.irec.tot_msgs_sent, int, 4)
        self._dbstmnt.bind_in("tot_msgs_failed", self.irec.tot_msgs_failed, int, 4)
        self._dbstmnt.bind_in("tot_msgs_retrys", self.irec.tot_msgs_retrys, int, 4)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.msg_req import tMsgReq

class dMsgReqInsert:

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tMsgReq()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    async def exec_deft(self,
                        id: int,
                        notype_id: str,
                        priority: int,
                        status: str,
                        time_req: datetime.datetime,
                        time_to_send: datetime.datetime,
                        time_proc: datetime.datetime,
                        msg_addr: dict,
                        msg_dict: dict,
                        cust_res: bool,
                        meta_data: dict,
                        modified_by: str,
                        tm_stamp: datetime.datetime) -> "dMsgReqInsert":
        """
        Execute the query by setting all the inputs.

        :param id: int
        :param notype_id: str
        :param priority: int
        :param status: str
        :param time_req: datetime.datetime
        :param time_to_send: datetime.datetime
        :param time_proc: datetime.datetime
        :param msg_addr: dict
        :param msg_dict: dict
        :param cust_res: bool
        :param meta_data: dict
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        :return: Self for convenience.
        """
        self.irec.id = id
        self.irec.notype_id = notype_id
        self.irec.priority = priority
        self.irec.status = status
        self.irec.time_req = time_req
        self.irec.time_to_send = time_to_send
        self.irec.time_proc = time_proc
        self.irec.msg_addr = msg_addr
        self.irec.msg_dict = msg_dict
        self.irec.cust_res = cust_res
        self.irec.meta_data = meta_data
        self.irec.modified_by = modified_by
        self.irec.tm_stamp = tm_stamp

        return await self.exec()

    async def exec(self, irec: tMsgReq = None) -> "dMsgReqInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = await self._dbcon.statement("MsgReqInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into loco.MsgReq (
  notype_id,
  priority,
  status,
  time_req,
  time_to_send,
  time_proc,
  msg_addr,
  msg_dict,
  cust_res,
  meta_data,
  modified_by,
  tm_stamp
) values (
  :notype_id,
  :priority,
  :status,
  :time_req,
  :time_to_send,
  :time_proc,
  :msg_addr,
  :msg_dict,
  :cust_res,
  :meta_data,
  :modified_by,
  :tm_stamp
)""")

        self.irec.tm_stamp = await self._dbcon.get_timestamp()

        self._dbstmnt.bind_in("notype_id", self.irec.notype_id, str, 128)
        self._dbstmnt.bind_in("priority", self.irec.priority, int, 4)
        self._dbstmnt.bind_in("status", self.irec.status, str, 1)
        self._dbstmnt.bind_in("time_req", self.irec.time_req, datetime.datetime)
        self._dbstmnt.bind_in("time_to_send", self.irec.time_to_send, datetime.datetime)
        self._dbstmnt.bind_in("time_proc", self.irec.time_proc, datetime.datetime)
        self._dbstmnt.bind_in("msg_addr", self.irec.msg_addr, dict)
        self._dbstmnt.bind_in("msg_dict", self.irec.msg_dict, dict)
        self._dbstmnt.bind_in("cust_res", self.irec.cust_res, bool)
        self._dbstmnt.bind_in("meta_data", self.irec.meta_data, dict)
        self._dbstmnt.bind_in("modified_by", self.irec.modified_by, str, 128)
        self._dbstmnt.bind_in("tm_stamp", self.irec.tm_stamp, datetime.datetime)

        await self._dbcon.execute(self._dbstmnt)

        self.irec.id = await self._dbcon.post_get_sequence(self._dbstmnt, "loco.MsgReq", "id", 8)

        if irec:
            irec._copy_from(self.irec)

        return self




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#

import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from .snap_prov_key import tSnapProvKey

class tSnapProv(mettle.io.ISerializable, mettle.db.IPrimaryKey):

    def __init__(self,
                 valuedate: datetime.date = datetime.date.min,
                 corrtype_id: str = '',
                 corrprov_id: str = '',
                 msg_sent: int = 0,
                 msg_failed: int = 0,
                 msg_retry: int = 0,
                 avg_msg_time: float = 0.0,
                 tot_msg_time: float = 0.0):
        """
        Constructor.

        :param valuedate: datetime.date
        :param corrtype_id: str
        :param corrprov_id: str
        :param msg_sent: int
        :param msg_failed: int
        :param msg_retry: int
        :param avg_msg_time: float
        :param tot_msg_time: float
        """
        self.valuedate = valuedate
        self.corrtype_id = corrtype_id
        self.corrprov_id = corrprov_id
        self.msg_sent = msg_sent
        self.msg_failed = msg_failed
        self.msg_retry = msg_retry
        self.avg_msg_time = avg_msg_time
        self.tot_msg_time = tot_msg_time

    def clear(self):
        """
        Clears all member variables for this database record.
        """
        self.valuedate = datetime.date.min
        self.corrtype_id = ''
        self.corrprov_id = ''
        self.msg_sent = 0
        self.msg_failed = 0
        self.msg_retry = 0
        self.avg_msg_time = 0.0
        self.tot_msg_time = 0.0

    def _name(self) -> str:
        """
        Name of the record.

        :return: The name.
        """
        return 'tSnapProv'

    def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
        """
        Serialize record to a stream.

        :param _w:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _w.write_start(_oname)
        _w.write_date("valuedate", self.valuedate)
        _w.write_string("corrtype_id", self.corrtype_id)
        _w.write_string("corrprov_id", self.corrprov_id)
        _w.write_int32("msg_sent", self.msg_sent)
        _w.write_int32("msg_failed", self.msg_failed)
        _w.write_int32("msg_retry", self.msg_retry)
        _w.write_double("avg_msg_time", self.avg_msg_time)
        _w.write_double("tot_msg_time", self.tot_msg_time)
        _w.write_end(_oname)

    def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
        """
        Deserialize record from a stream.

        :param _r:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _r.read_start(_oname)
        self.valuedate = _r.read_date("valuedate")
        self.corrtype_id = _r.read_string("corrtype_id")
        self.corrprov_id = _r.read_string("corrprov_id")
        self.msg_sent = _r.read_int32("msg_sent")
        self.msg_failed = _r.read_int32("msg_failed")
        self.msg_retry = _r.read_int32("msg_retry")
        self.avg_msg_time = _r.read_double("avg_msg_time")
        self.tot_msg_time = _r.read_double("tot_msg_time")
        _r.read_end(_oname)

    def _copy_from(self, rec: "tSnapProv"):
        """
        Copies the column from the rec into this record.

        :param rec: The source record.
        """
        self.valuedate = rec.valuedate
        self.corrtype_id = rec.corrtype_id
        self.corrprov_id = rec.corrprov_id
        self.msg_sent = rec.msg_sent
        self.msg_failed = rec.msg_failed
        self.msg_retry = rec.msg_retry
        self.avg_msg_time = rec.avg_msg_time
        self.tot_msg_time = rec.tot_msg_time

    def _write_pk(self, _key: "tSnapProv"):
        """
        Write the primary key columns to the key object.

        :param _key:
        """
        self._write_key(_key)

    def _read_pk(self, _key: "tSnapProv"):
        """
        Read the primary key columns from the key object.

        :param _key:
        """
        self._read_key(_key)

    def _serialize_key(self, _w: mettle.io.IWriter):
        """
        Serialize the primary key to a stream.

        :param _w:
        """
        _key = tSnapProvKey()

        self._write_key(_key)
        _key._serialize(_w)

    def _deserialize_key(self, _r: mettle.io.IReader):
        """
        Deserialize the primary key from a stream.

        :param _r:
        """
        _key = tSnapProvKey()

        _key._deserialize(_r)
        self._read_key(_key)

    @staticmethod
    def _cache_davs(dvc=None, for_ins: bool = False, key_only: bool = False):
        """
        Cache the DAV into the dav cache (or create one) and return it.

        :param dvc: (mettle.lib.DavCache), target cache, if None a new one is created and returned.
        :param for_ins: Davs for insert.
        :param key_only: Davs by key only, ie for delete/selectone.
        """
        if dvc is None:
            dvc = mettle.lib.DavCache()

        dvc.add_targ("valuedate", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("corrtype_id", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("corrtype_id", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 16))
        dvc.add_targ("corrprov_id", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("corrprov_id", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 16))
        if key_only:
            return dvc

        dvc.add_targ("msg_sent", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("msg_failed", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("msg_retry", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("avg_msgTime", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("tot_msgTime", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))

        return dvc

    def _get_davs(self, dvc=None, for_ins: bool = False, key_only: bool = False):
        tSnapProv._cache_davs(dvc, for_ins, key_only)

    def _write_key(self, _key: tSnapProvKey):
        """
        Write primary key columns to _key.

        :param _key:
        """
        _key.valuedate = self.valuedate
        _key.corrtype_id = self.corrtype_id
        _key.corrprov_id = self.corrprov_id

    def _read_key(self, _key: tSnapProvKey):
        """
        Read the primary key columns from _key.

        :param _key:
        """
        self.valuedate = _key.valuedate
        self.corrtype_id = _key.corrtype_id
        self.corrprov_id = _key.corrprov_id

    @staticmethod
    def _pk_compare(x, y) -> int:
        """
        Compare function do primary key comparisons for list sorts and searches.

        :return: zero for x == y, negative for x < y, positive for x > y
        """
        rc = (x.valuedate > y.valuedate) - (x.valuedate < y.valuedate)

        if rc != 0:
            return rc

        rc = (x.corrtype_id > y.corrtype_id) - (x.corrtype_id < y.corrtype_id)

        if rc != 0:
            return rc

        rc = (x.corrprov_id > y.corrprov_id) - (x.corrprov_id < y.corrprov_id)

        return rc

    def __lt__(self, other):
        return other and self._pk_compare(self, other) < 0

    def __gt__(self, other):
        return other and self._pk_compare(self, other) > 0

    def __eq__(self, other):
        return other and self._pk_compare(self, other) == 0

    def __le__(self, other):
        return other and self._pk_compare(self, other) <= 0

    def __ge__(self, other):
        return other and self._pk_compare(self, other) >= 0

    def __ne__(self, other):
        return other == None or self._pk_compare(self, other)

    def __repr__(self) -> str:
        return '<SnapProv [valuedate:%s, corrtype_id:%.32s, corrprov_id:%.32s, msg_sent:%r, msg_failed:%r, msg_retry:%r, avg_msg_time:%r, tot_msg_time:%r]>' % (
            self.valuedate, self.corrtype_id, self.corrprov_id, self.msg_sent, self.msg_failed, self.msg_retry, self.avg_msg_time, self.tot_msg_time)


    class List(list, mettle.io.ISerializable):
        """
        List Class
        """

        def _name(self) -> str:
            """
            Name of the list.

            :return: The name
            """
            return 'tSnapProv.List'

        def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
            """
            Serialize the list to a stream.

            :param _w:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _w.write_start_list(_oname, len(self))

            for _rec in self:
                _rec._serialize(_w)

            _w.write_end(_oname)

        def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
            """
            Deserialize the list from a stream.

            :param _r:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _cnt = _r.read_start_list(_oname)

            while _cnt >= 1:
                _rec  = tSnapProv()
                _cnt -= 1
                _rec._deserialize(_r)
                self.append(_rec)

            _r.read_end(_oname)

        def __repr__(self) -> str:
            return '<tSnapProv.List[%d]>' % (len(self))

        def _get_davs(self, dvc=None, for_ins: bool = False, key_only: bool = False):
            return tSnapProv._cache_davs(dvc, for_ins, key_only)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#

import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

class oMsgReqPendingChanged(mettle.io.ISerializable):

    def __init__(self,
                 msg_req_id: int = 0,
                 msg_status: str = '',
                 msg_time: datetime.datetime = datetime.datetime.min):
        """
        Constructor.

        :param msg_req_id: int
        :param msg_status: str
        :param msg_time: datetime.datetime
        """
        self.msg_req_id = msg_req_id
        self.msg_status = msg_status
        self.msg_time = msg_time

    def clear(self):
        """
        Clears all member variables for this database record.
        """
        self.msg_req_id = 0
        self.msg_status = ''
        self.msg_time = datetime.datetime.min

    def _name(self) -> str:
        """
        Name of the record.

        :return: The name.
        """
        return 'oMsgReqPendingChanged'

    def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
        """
        Serialize record to a stream.

        :param _w:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _w.write_start(_oname)
        _w.write_int64("msg_req_id", self.msg_req_id)
        _w.write_char("msg_status", self.msg_status)
        _w.write_datetime("msg_time", self.msg_time)
        _w.write_end(_oname)

    def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
        """
        Deserialize record from a stream.

        :param _r:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _r.read_start(_oname)
        self.msg_req_id = _r.read_int64("msg_req_id")
        self.msg_status = _r.read_char("msg_status")
        self.msg_time = _r.read_datetime("msg_time")
        _r.read_end(_oname)

    def _copy_from(self, rec: "oMsgReqPendingChanged"):
        """
        Copies the column from the rec into this record.

        :param rec: The source record.
        """
        self.msg_req_id = rec.msg_req_id
        self.msg_status = rec.msg_status
        self.msg_time = rec.msg_time

    def __repr__(self) -> str:
        return '<MsgReq [msg_req_id:%r, msg_status:%.32s, msg_time:%s]>' % (
            self.msg_req_id, self.msg_status, self.msg_time)


    class List(list, mettle.io.ISerializable):
        """
        List Class
        """

        def _name(self) -> str:
            """
            Name of the list.

            :return: The name
            """
            return 'oMsgReqPendingChanged.List'

        def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
            """
            Serialize the list to a stream.

            :param _w:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _w.write_start_list(_oname, len(self))

            for _rec in self:
                _rec._serialize(_w)

            _w.write_end(_oname)

        def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
            """
            Deserialize the list from a stream.

            :param _r:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _cnt = _r.read_start_list(_oname)

            while _cnt >= 1:
                _rec  = oMsgReqPendingChanged()
                _cnt -= 1
                _rec._deserialize(_r)
                self.append(_rec)

            _r.read_end(_oname)

        def __repr__(self) -> str:
            return '<oMsgReqPendingChanged.List[%d]>' % (len(self))




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #

"""
LOCO Template upload/download batch class.
"""

import os.path
import json
import yaml

from bs_lib  import Pod
from bs_lib  import common

import bs_loco

from bs_loco.db.tables import tNoTempl


class BatchUploadDownloadTemplate:
    """
    Provides a class that can be easily used to upload or download correspondence templates files into and
    out of the database.
    """

    def __init__(self, pod: Pod, dao_name: str):
        """
        Constructor

        :param pod: The pod to use.
        :param dao_name: The dao objects to use.
        """
        self._pod = pod
        self._log = pod.log
        self._dao = bs_loco.dao_by_name(dao_name)


    def download_template(self, templ_id: str, fmt: str = 'yaml', dest_filename: str = None) -> None:
        """
        Download the loco template to a yaml file.

        :param templ_name: The template id to download (see the `notempl` table).
        :param fmt: Download the file in either `json` or `yaml` format. The default is `yaml`.
        :param dest_filename: Optionally specifiy the destination file name, if left empty uses the the
                              template name.
        """
        notype    = self._dao.dNoType(self._pod.dbcon)
        templ_qry = self._dao.dNoTemplByNoType(self._pod.dbcon)

        if not notype.try_select_one_deft(templ_id):
            raise Exception('Loco Type not found. [id:%s]' % (templ_id))

        out_pl = {
            'id'  : templ_id,
        }

        templ_qry.exec_deft(templ_id, '')

        while templ_qry.fetch():
            if not templ_qry.orec.templ:
                continue

            out_pl[templ_qry.orec.corrtype_id] = templ_qry.orec.templ


        if dest_filename:
            fname = dest_filename
        else:
            fname = f'{templ_id}.{fmt}'

        with open(fname, 'wt') as fh:
            if fmt == 'json':
                fh.write(json.dumps(out_pl))
            else:
                fh.write(yaml.safe_dump(out_pl))

        self._log.info(f' - downloaded template [file: {fname}] ok')


    def upload_template(self, path_to_templ: str) -> None:
        """
        Upload a yaml or json template file.

        :param path_to_templ: The path to the json or yaml template file to upload.
        """
        if not os.path.exists(path_to_templ):
            raise Exception('File not found [%s]' % (path_to_templ))

        data      = common.read_config_dict_from_file(path_to_templ)
        templ_id  = common.read_dict(data, 'id',    str)
        notype    = self._dao.dNoType(self._pod.dbcon)
        ct_qry    = self._dao.dCorrProvSelectAll(self._pod.dbcon)
        no_templ  = self._dao.dNoTemplBy(self._pod.dbcon)
        mod_by    = '[bs_loco.BatchUploadDownloadTemplate]'
        ucnt      = 0

        if not notype.try_select_one_deft(templ_id):
            raise Exception('Loco Type not found. [id:%s]' % (templ_id))

        ct_qry.exec()

        while ct_qry.fetch():
            ct_data = common.read_dict(data, ct_qry.orec.id, dict, optional=True)

            # TODO, validate the schema of this dictionary, will need to pull in cerebrus parser.

            if not ct_data:
                continue

            # TODO, update the `notemplcfg` table with the old schema if it changed

            if no_templ.try_select_one_deft(templ_id, ct_qry.orec.id):
                no_templ.rec.status      = tNoTempl.Status_Couplet.key_active
                no_templ.rec.templ       = ct_data
                no_templ.rec.modified_by = mod_by
                no_templ.update()

                self._log.info(f' - uploaded template [notype_id: {no_templ.rec.notype_id},'
                               ' corrtype_id: {no_templ.rec.corrtype_id}] ok')

            else:
                no_templ.insert_deft(templ_id,
                                     ct_qry.orec.id,
                                     tNoTempl.Status_Couplet.key_active,
                                     None,
                                     ct_data,
                                     None,
                                     mod_by)

                self._log.info(f' - inserted template [notype_id: {no_templ.rec.notype_id},'
                               ' corrtype_id: {no_templ.rec.corrtype_id}] ok')

            ucnt += 1

        if not ucnt:
            self.log.warning(' - no valid correspondence types configured for upload in file, aborting.')
            return

        self._log.info(f' - uploaded template [file: {path_to_templ}] ok')

        self._pod.dbcon.commit()



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#

import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

class iMsgReqUpdateStatus(mettle.io.ISerializable):

    def __init__(self,
                 id: int = 0,
                 status: str = '',
                 time_proc: datetime.datetime = datetime.datetime.min,
                 modified_by: str = ''):
        """
        Constructor.

        :param id: int
        :param status: str
        :param time_proc: datetime.datetime
        :param modified_by: str
        """
        self.id = id
        self.status = status
        self.time_proc = time_proc
        self.modified_by = modified_by

    def clear(self):
        """
        Clears all member variables for this database record.
        """
        self.id = 0
        self.status = ''
        self.time_proc = datetime.datetime.min
        self.modified_by = ''

    def _name(self) -> str:
        """
        Name of the record.

        :return: The name.
        """
        return 'iMsgReqUpdateStatus'

    def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
        """
        Serialize record to a stream.

        :param _w:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _w.write_start(_oname)
        _w.write_int64("id", self.id)
        _w.write_char("status", self.status)
        _w.write_datetime("time_proc", self.time_proc)
        _w.write_string("modified_by", self.modified_by)
        _w.write_end(_oname)

    def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
        """
        Deserialize record from a stream.

        :param _r:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _r.read_start(_oname)
        self.id = _r.read_int64("id")
        self.status = _r.read_char("status")
        self.time_proc = _r.read_datetime("time_proc")
        self.modified_by = _r.read_string("modified_by")
        _r.read_end(_oname)

    def _copy_from(self, rec: "iMsgReqUpdateStatus"):
        """
        Copies the column from the rec into this record.

        :param rec: The source record.
        """
        self.id = rec.id
        self.status = rec.status
        self.time_proc = rec.time_proc
        self.modified_by = rec.modified_by

    @staticmethod
    def _cache_davs(dvc=None):
        """
        Cache the DAV into the dav cache (or create one) and return it.

        :param dvc: (mettle.lib.DavCache), target cache, if None a new one is created and returned.
        """
        if dvc is None:
            dvc = mettle.lib.DavCache()

        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.Min, 1))
        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.Min, 1))
        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("status", mettle.lib.Dav(mettle.lib.Dav.eDavType.Min, 1))
        dvc.add_targ("status", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 1))
        dvc.add_targ("status", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("status", mettle.lib.Dav(mettle.lib.Dav.eDavType.In, ['P','B','R','C']))
        dvc.add_targ("modified_by", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("modified_by", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 128))

        return dvc

    def _get_davs(self, dvc=None):
        iMsgReqUpdateStatus._cache_davs(dvc)

    def __repr__(self) -> str:
        return '<MsgReq [id:%r, status:%.32s, time_proc:%s, modified_by:%.32s]>' % (
            self.id, self.status, self.time_proc, self.modified_by)


    class List(list, mettle.io.ISerializable):
        """
        List Class
        """

        def _name(self) -> str:
            """
            Name of the list.

            :return: The name
            """
            return 'iMsgReqUpdateStatus.List'

        def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
            """
            Serialize the list to a stream.

            :param _w:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _w.write_start_list(_oname, len(self))

            for _rec in self:
                _rec._serialize(_w)

            _w.write_end(_oname)

        def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
            """
            Deserialize the list from a stream.

            :param _r:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _cnt = _r.read_start_list(_oname)

            while _cnt >= 1:
                _rec  = iMsgReqUpdateStatus()
                _cnt -= 1
                _rec._deserialize(_r)
                self.append(_rec)

            _r.read_end(_oname)

        def __repr__(self) -> str:
            return '<iMsgReqUpdateStatus.List[%d]>' % (len(self))

        def _get_davs(self, dvc=None):
            return iMsgReqUpdateStatus._cache_davs(dvc)




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import datetime
import smtplib
import os.path

import email.encoders
import email.mime.base
import email.mime.multipart
import email.mime.text
import email.mime.image

from bs_lib.Pod   import Pod

from bs_loco.db   import tMsg

from .provider_email import ProviderEmail

from bs_lib  import Common


class SmtpEmail(ProviderEmail):
    """
    The standard SMTP email provider.
    """

    def __init__(self):
        """
        Constructor.
        """
        ProviderEmail.__init__(self)
        self._smtp = None


    def id(self) -> str:
        """
        Overload.
        """
        return 'smtp'


    def destroy(self):
        """!
        Cleans up the provider.
        """
        self._logoff()
        ProviderEmail.destroy(self)


    def initialize(self, pod: Pod, dao: object, cfg: dict):
        """
        Overload.
        """
        ProviderEmail.initialize(self, pod, dao, cfg)

        Common.readDict(self._cfg, 'host',        str)
        Common.readDict(self._cfg, 'port',        int)
        Common.readDict(self._cfg, 'username',    str)
        Common.readDict(self._cfg, 'passwd',      str)
        Common.readDict(self._cfg, 'from',        str)
        Common.readDict(self._cfg, 'ssl',         bool)
        Common.readDict(self._cfg, 'tls',         bool)
        Common.readDict(self._cfg, 'timeout',     float)


    def _logoff(self):
        """
        Log off of the smtp server.
        """
        if self._smtp:
            try:
                self._smtp.quit()
            except smtplib.SMTPServerDisconnected as x:
                self._log.warning('Smtp already disconnected in destroy: %s' % str(x))
            finally:
                self._smtp = None


    def _login(self):
        """
        Log onto the smtp server.
        """
        self._logoff()

        if self._cfg.get('ssl'):
            self._smtp = smtplib.SMTP_SSL(self._cfg['host'], self._cfg['port'], timeout=self._cfg['timeout'])
        else:
            self._smtp = smtplib.SMTP(self._cfg['host'], self._cfg['port'], self._cfg['timeout'])

        if self._cfg.get('tls'):
            self._smtp.starttls()

        # ensure we can login
        if len(self._cfg['username']) < 1:
            raise Exception('Cannot log into smtp server without a user name!')

        self._smtp.login(os.path.expandvars(self._cfg['username']), os.path.expandvars(self._cfg['passwd']))

        self._log.debug('smtp initialized [host:%s, user:%s]' % (self._cfg['host'], self._cfg['username']))


    def _provider_send(self, payload: dict, res_list: list, msg: tMsg, addr_list: list) -> tuple:
        """
        Overload.
        """
        try:
            dtstart = datetime.datetime.now()

            self._login()

            emsg = email.mime.multipart.MIMEMultipart('related')

            emsg['Subject'] = payload['subject']
            emsg['From']    = self._cfg['from']
            emsg['To']      = ';'.join(addr_list)
            emsg['Date']    = email.utils.formatdate(localtime=True)
            emsg.preamble   = 'This is a multi-part message in MIME format.'

            msg_alt = email.mime.multipart.MIMEMultipart('alternative')
            emsg.attach(msg_alt)

            msg_alt.attach(email.mime.text.MIMEText(payload['body-html'], 'html'))
            msg_alt.attach(email.mime.text.MIMEText(payload['body-text']))

            if res_list:
                for att in res_list:
                    if not att.content or not att.mimetype or not att.mimetype.startswith('image/'):
                        continue

                    att_img = email.mime.image.MIMEImage(att.content)
                    att_img.add_header('Content-ID', '<{}>'.format(att.id))
                    emsg.attach(att_img)

                for att in res_list:
                    if not att.content or (att.mimetype and att.mimetype.startswith('image/')):
                        continue

                mtype             = att.mimetype or 'application/octet-stream'
                maintype, subtype = mtype.split('/', 1)

                part = email.mime.base.MIMEBase(maintype, subtype)
                part.set_payload(att.content)

                email.encoders.encode_base64(part)

                part.add_header('Content-Disposition', 'attachment', filename = att.filename)
                emsg.attach(part)

            self._smtp.sendmail(self._cfg['from'], addr_list, emsg.as_string())

            dtend = datetime.datetime.now()

            self._log.debug(' - mail sent successfull')
        except Exception as x:
            self._log.error('Exception caught sending email [%s]' % (str(msg)))
            self._log.exception(x)
            return (False, None, str(x))
        finally:
            self._logoff()

        return (True, dtend - dtstart, None)



# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #

from bs_lib import Pod, PodAsync

from bs_loco           import constants
from bs_loco.xloco     import xLoco

from .loco_trig      import LocoTrig
from .local_db       import LocalDB
from .stub           import Stub

from .loco_trig_async  import LocoTrigAsync
from .local_db_async   import LocalDBAsync
from .stub_async       import StubAsync


def get_trigger(pod: Pod, trig_id: str, cfg: dict = None) -> LocoTrig:
    """
    Factory to load a loco trigger by id.

    :param pod: The pod to use.
    :param trig_id: Loco trigger identifier.
    :param cfg: The trigger config object
    """
    res = None

    if trig_id == constants.LOCO_TRIGGER_LOCAL_DB:
        res = LocalDB()
    elif trig_id == constants.LOCO_TRIGGER_STUB:
        res = Stub()
    # TODO, add kafka queue or other mechanisms here

    if not res:
        raise xLoco('Loco Trigger [%s] not expected.' % trig_id)

    res.initialize(pod, cfg)

    if res.is_deadbeef():
        del res
        res = Stub()
        res.initialize(pod, cfg)

    return res


async def get_trigger_async(pod: PodAsync, trig_id: str, cfg: dict = None) -> LocoTrigAsync:
    """
    Factory to load am async loco trigger by id.

    :param pod: The pod to use.
    :param trig_id: Loco trigger identifier.
    :param cfg: The trigger config object
    """
    res = None

    if trig_id == constants.LOCO_TRIGGER_LOCAL_DB:
        res = LocalDBAsync()
    elif trig_id == constants.LOCO_TRIGGER_STUB:
        res = StubAsync()
    # TODO, add kafka queue or other mechanisms here

    if not res:
        raise xLoco('Loco Trigger [%s] not expected.' % trig_id)

    await res.initialize(pod, cfg)

    if res.is_deadbeef():
        del res
        res = StubAsync()
        await res.initialize(pod, cfg)

    return res



# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #

import datetime
import time
import traceback

from bs_lib.auto_transaction import AutoTransaction
from bs_lib                  import common
from bs_lib                  import query_builder

import bs_loco

from bs_loco.config  import Config

from bs_loco import corrprov


class LocoServer:
    """
    This server monitor for new messages requests processes them to their
    various providers.
    """
    MODULE                = '[LocoServer]'
    PROC_INTERVAL_SLEEP   = 0.2
    MAX_PIPE_OUT_LEN      = 4096
    SCHEDULE_INTERVAL     = 10.0
    MSG_PROC_PER_INTERVAL = 1000

    PROC_TYPE_SEND_MSG    = 'loco-send-msg'


    def __init__(self, pod, dao_name: str, corrtype_list: list):
        """
        Constructor.

        :param pod: The pod to use.
        :param dao_name: The data access object library to use.
        :param corrtype_list: List of correspodence types to process for.
        """
        self._shutdown      = False
        self._pod           = pod
        self._log           = pod.log
        self._dao           = bs_loco.dao_by_name(dao_name)
        self._corrtype_list = corrtype_list
        self._prov_cache    = { '__cache__' : datetime.datetime.min }

        self._proc_interval   = 1.0
        self._prov_interval   = 60.0 * 60.0  # 1 hour


    def shutdown(self):
        """
        Instruct the server to shutdown.
        """
        self._shutdown = True


    def run(self):
        """
        Run the server.
        """
        self._shutdown  = False

        try:
            self._initialize()
            self._cache_providers(datetime.datetime.now())

            self._log.info("Loco Server - starting [procInterval:%.2f]" % (self._proc_interval))

            while not self._shutdown:
                self._process(datetime.datetime.now())

                if self._shutdown:
                    break

                pi = self._proc_interval

                while pi > 0.0:
                    pi -= self._proc_interval
                    time.sleep(self._proc_interval)

                    if self._shutdown:
                        break

            self._log.info("Loco Server - stopping")

        except Exception as x:
            self._log.error('Loco Server exception caught [error:%s, trace:%s]' % (str(x), traceback.format_exc()))
            raise
        finally:
            self._destroy()


    def run_once(self) -> int:
        """
        This is debug function that only runs the server loop once.

        :returns: The number of messages processed.
        """
        self._shutdown = False

        self._initialize()

        self._cache_providers(datetime.datetime.now())

        self._log.info("Loco Server - run_once start")

        cnt = self._process(datetime.datetime.now())

        self._shutdown = True

        self._destroy()

        self._log.info("Loco Server - run_once stop [cnt:%d]" % (cnt))

        return cnt


    def _initialize(self):
        """
        Initialize the generation server.
        """
        with AutoTransaction(self._pod):
            conf = Config(self._pod, self._dao)
            cfg  = conf.load([
                'server.proc-interval',
                'server.prov-interval',
            ])

            print(cfg)

            pi = common.read_dict(cfg, 'server.proc-interval', float)
            pv = common.read_dict(cfg, 'server.prov-interval', float)

            if pi < 0.2:
                self._log.warning('Interval value invalid [%.2f], using default of [%.2f]' % self._proc_interval)
            else:
                self._proc_interval = pi
                self._log.debug(' - server.proc-interval : %.2f' % self._proc_interval)

            if pv < 60.0:
                self._log.warning('Interval value invalid [%.2f], using default of [%.2f]' % self._prov_interval)
            else:
                self._prov_interval = pv
                self._log.debug(' - server.prov-interval : %.2f' % self._prov_interval)


    def _destroy(self):
        """
        Release any objects that are loaded.
        """
        pass


    def _process(self, dtnow: datetime.datetime) -> int:
        """
        Process any new messages.

        :param dtnow: To run the the specified time.
        :return: Number of messages processed.
        """
        msgs  = []
        idx   = 0

        with AutoTransaction(self._pod):
            self._dao.dMsgForProcessing(self._pod.dbcon).exec_deft(
                query_builder.dyn_list(self._corrtype_list, 'm', 'corrtype_id'),
                dtnow,
                self.MSG_PROC_PER_INTERVAL).fetch_all(msgs)

            if not msgs:
                return 0

            self._cache_providers(dtnow)

        self._log.info('Processing [%d] messages' % (len(msgs)))

        for msg in msgs:
            if self._shutdown:
                self._log.warning('Process shutdown detected at msg [idx:%d, id:%d], breaking loop.' % (idx, msg.id))

            if idx % 100 == 0:
                self._log.info(' ...[%d/%d]' % (idx, len(msgs)))

            if msg.corrtype_id not in self._prov_cache:
                raise Exception('Internal error, corrtype_id not found in provider cache!' % msg.corrtype_id)

            if not self._prov_cache[msg.corrtype_id].send(msg.id):
                self._log.warning("Msg failed to send, see provider log for detail. [%s]" % str(msg))

            idx += 1

        self._log.info(' ...[%d/%d]' % (idx, len(msgs)))

        return idx


    def _destroy_providers(self):
        """
        Destroys all the providers.
        """
        for pr in self._corrtype_list:
            if pr not in self._prov_cache:
                continue

            probj = self._prov_cache.pop(pr)
            probj.destroy()


    def _cache_providers(self, dtnow: datetime.datetime):
        """
        Cache the required providers.
        """
        if self._prov_cache['__cache__'] > dtnow:
            return

        self._destroy_providers()  # Note consider detecing for changes, and only reload the providers that have changed.

        with AutoTransaction(self._pod):
            for pr in self._corrtype_list:
                self._prov_cache[pr] = corrprov.get_provider(self._pod, self._dao, pr)

        self._prov_cache['__cache__'] = dtnow + datetime.timedelta(seconds = self._prov_interval)



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.no_res import tNoRes

class dNoResUpdate:

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tNoRes()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    def exec_deft(self,
                  id: str,
                  filename: str,
                  mimetype: str,
                  content: bytearray,
                  modified_by: str) -> "dNoResUpdate":
        """
        Execute the query by setting all the inputs.

        :param id: str
        :param filename: str
        :param mimetype: str
        :param content: bytearray
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        :return: Self for convenience.
        """
        self.irec.id = id
        self.irec.filename = filename
        self.irec.mimetype = mimetype
        self.irec.content = content
        self.irec.modified_by = modified_by

        return self.exec()

    def exec(self, irec: tNoRes = None) -> "dNoResUpdate":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("NoResUpdate", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""update
  loco.NoRes
set
  filename = :filename,
  mimetype = :mimetype,
  content = :content,
  modified_by = :modified_by,
  tm_stamp = :tm_stamp
where
  id = :id""")

        self.irec.tm_stamp = self._dbcon.get_timestamp()

        self._dbstmnt.bind_in("id", self.irec.id, str, 128)
        self._dbstmnt.bind_in("filename", self.irec.filename, str, 256)
        self._dbstmnt.bind_in("mimetype", self.irec.mimetype, str, 128)
        self._dbstmnt.bind_in("content", self.irec.content, bytearray)
        self._dbstmnt.bind_in("modified_by", self.irec.modified_by, str, 128)
        self._dbstmnt.bind_in("tm_stamp", self.irec.tm_stamp, datetime.datetime)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import datetime
import json
import re
import traceback

from bs_lib.AutoTransaction import AutoTransaction
from bs_lib.Pod             import Pod
from bs_lib                 import QueryBuilder

from bs_loco.loco_template  import LocoTemplate
from bs_loco.xloco          import xLoco

from bs_loco.db.tables   import tMsg
from bs_loco.db.tables   import tMsgReq
from bs_loco.db.tables   import tNoType


class Provider:
    """
    This is the base class for all correspondence types.
    """
    MAX_MSG_ERROR_LOGS = 5

    def __init__(self):
        """
        Constructor.
        """
        self._pod   = None
        self._dao   = None
        self._cfg   = None
        self._cache = {}


    def __del__(self):
        """
        Destructor.
        """
        self.destroy()


    def initialize(self, pod: Pod, dao: object, cfg: dict):
        """
        Virtual method to initialize the loco trigger interface.

        :param pod: The pod to use.
        :param dao: The dao to use.
        :param cfg: Any configs needed for the trigger, can be None.
        """
        self.destroy()
        self._pod = pod
        self._dao = dao
        self._cfg = cfg
        self._log = pod.log

        self._cache = {
            'notype'  : {},
            'templ'   : {},
            'res'     : {},
            'keypair' : None,
            'allow'   : None,
        }

        dcp = dao.dCorrProv(pod.dbcon)

        dcp.select_one_deft(self.id(), self.corrtype_id())


    def destroy(self):
        """
        Virtual method to destroy the content management object.
        """
        self._pod = None
        self._dao = None
        self._cfg = None
        self._log = None
        self._cache.clear()


    def id(self) -> str:
        """
        Pure virtual method, the identifier of the content management overload.

        :return: The overload name.
        """
        raise Exception('id() not implemented.')


    def corrtype_id(self) -> str:
        """
        Pure virtual method, the identifier of the content management overload.

        :return: The overload name.
        """
        raise Exception('corrtype_id() not implemented.')


    def name(self) -> str:
        """
        Pure virtual method, the name of the content management overload.

        :return: The overload name.
        """
        return '%s.%s' % (self.corrtype_id(), self.id())


    def is_deadbeef(self) -> bool:
        """
        Virtual method to check if the interface is active but not configured.

        :return: True if this interface is active but not configure.
        """
        return False


    def send(self, msg_id) -> bool:
        """
        Send a message to a provider.

        :param msg_id: The message to send.
        :return: True of success or false if something went wrong.
        """
        self._log.debug('send - start [msgId:%d]' % (msg_id))

        dmsg = self._dao.dMsg(self._pod.dbcon)
        dreq = self._dao.dMsgReq(self._pod.dbcon)

        with AutoTransaction(self._pod) as at:
            try:
                dmsg.lock_one_deft(msg_id, self._pod.stdDBLock())

                if self.corrtype_id() != dmsg.rec.corrtype_id:
                    raise Exception('Internal error, provider received wrong message type. [corrtype_id:%s, msg:%s]' % (
                        self.corrtype_id(), str(dmsg.rec)))


                if dmsg.rec.status != tMsg.Status_Couplet.key_pending and\
                   dmsg.rec.status != tMsg.Status_Couplet.key_retry:
                    self._log.warning('  - skipping msg in unexepcted status [status:%s]' % (dmsg.rec.status))
                    return True

                if dmsg.rec.status == tMsg.Status_Couplet.key_retry:
                    dmsg.rec.retry_cnt += 1

                ok, err_msg = self._init_cache(dmsg.rec)

                if not ok:
                    dmsg.rec.status    = tMsg.Status_Couplet.key_template_error
                    dmsg.rec.msglog    = self._push_msg_log(dmsg.rec.msglog, err_msg)
                    dmsg.rec.fail_cnt += 1
                    dmsg.update()
                    at.commit()
                    return True

                ntrec = self._cache['notype'][dmsg.rec.notype_id]

                if ntrec.status != tNoType.Status_Couplet.key_active:
                    self._log.warning('  - skipping message, because the notification type is not active: [msg:%s]' % (
                        str(dmsg.rec)))
                    dmsg.rec.status = tMsg.Status_Couplet.key_cancelled
                    dmsg.rec.msglog = self._push_msg_log(dmsg.rec.msglog, 'Notification type is disabled')
                    dmsg.update()
                    at.commit()
                    return True

                dreq.select_one_deft(dmsg.rec.msgreq_id)

                if ntrec.max_life > 0:
                    dtnow = datetime.datetime.now()

                    if dreq.rec.time_to_send + datetime.timedelta(minutes=ntrec.max_life) < dtnow:
                        self._log.warning('  - skipping message, max life reached: [max_life:%d, now:%s, req_time:%s, msg:%s]' % (  # noqa
                            ntrec.max_life, str(dtnow), str(dreq.rec.time_to_send), str(dmsg.rec)))
                        dmsg.rec.status = tMsg.Status_Couplet.key_cancelled
                        dmsg.rec.msglog = self._push_msg_log(dmsg.rec.msglog, 'Max life reached [%d min]' % (ntrec.max_life))
                        dmsg.update()
                        at.commit()
                        return False

                try:
                    if dmsg.rec.corrprov_id == self.id() and dmsg.rec.payload:
                        payload = json.loads(dmsg.rec.payload)
                    else:
                        payload          = self._build_payload(dreq.rec)
                        dmsg.rec.payload = json.dumps(payload)

                    resource_list = self._load_msgreq_resources(dreq.rec)
                except Exception:

                    tb = traceback.format_exc()

                    self._log.error('Exception building payload or loading resources [msg_id:%d, provider:%s]' % (
                        msg_id, self.id()))

                    self._log.error(tb)

                    dmsg.rec.status    = tMsg.Status_Couplet.key_template_error
                    dmsg.rec.msglog    = self._push_msg_log(dmsg.rec.msglog, tb)
                    dmsg.rec.fail_cnt += 1
                    dmsg.update()
                    at.commit()
                    return False

                msg_addr = json.loads(dreq.rec.msg_addr)
                addr_list = self._addresses_allowed(msg_addr.get(self.corrtype_id()))

                if not addr_list:
                    self._log.warning('  - skipping message, because none of the addresses match\
 the allowed [msg:%s, addresses:%s]' % (str(dmsg.rec), str(msg_addr.get(self.corrtype_id()))))
                    dmsg.rec.status = tMsg.Status_Couplet.key_cancelled
                    dmsg.rec.msglog = self._push_msg_log(dmsg.rec.msglog, 'No addresses are allowed.')
                    dmsg.update()
                    at.commit()
                    return True

                ok, time_taken, errMsg = self._provider_send(payload,
                                                             resource_list,
                                                             dmsg.rec,
                                                             addr_list)

                dmsg.rec.corrprov_id = self.id()
                dmsg.rec.usrId       = self._pod.usr_id

                if ok:
                    dmsg.rec.time_sent  = datetime.datetime.now()
                    dmsg.rec.time_taken = time_taken.total_seconds()
                    dmsg.rec.status     = tMsg.Status_Couplet.key_sent
                else:
                    dmsg.rec.msglog    = self._push_msg_log(dmsg.rec.msglog, err_msg)
                    dmsg.rec.fail_cnt += 1
                    dmsg.rec.status    = tMsg.Status_Couplet.key_failed

                dmsg.update()

                self._upd_parent_status(dreq, dreq.rec)

                at.commit()

            except Exception as x:
                self._log.error('Exception caught sending [msg_id:%d, provider:%s]' % (msg_id, self.id()))
                self._log.exception(x)
                raise

        self._log.debug('send - done [msgId:%d, status:%s]' % (dmsg.rec.id, tMsg.Status_Couplet.get_value(dmsg.rec.status)))

        return dmsg.rec.status == tMsg.Status_Couplet.key_sent


    def _upd_parent_status(self, msg: tMsg, req: tMsgReq):
        """
        Updates the parent (MsgReq) table status.

        :param msg: The message record.
        :param msgreq: The message request parent.
        """
        if req.Status_Couplet == tMsgReq.Status_Couplet.key_processed:
            return

        lst        = []
        new_status = tMsgReq.Status_Couplet.key_busy

        self._dao.dMsgStatusCntByReq(self._pod.dbcon).exec_deft(req.id).fetch_all(lst)

        if len(lst) == 1:
            if lst[0].status == tMsg.Status_Couplet.key_cancelled:
                new_status = tMsgReq.Status_Couplet.key_cancelled
            elif lst[0].status == tMsg.Status_Couplet.key_sent:
                new_status = tMsgReq.Status_Couplet.key_processed
        else:
            all_done = True

            for item in lst:
                if item.status == tMsg.Status_Couplet.key_sent or item.status == tMsg.Status_Couplet.key_cancelled:
                    continue

                all_done = False
                break

            if all_done:
                new_status = tMsgReq.Status_Couplet.key_processed

        self._dao.dMsgReqUpdateStatus(self._pod.dbcon).exec_deft(req.id,
                                                                 new_status,
                                                                 datetime.datetime.now(),
                                                                 self._pod.usr_id)


    def _init_cache(self, msg: tMsg) -> tuple:
        """
        Loads all the objects needed for the message cache.

        :param msg: The message record.
        :returns: (bool succes_or_failure, str error_msg)
        """
        try:
            if msg.notype_id not in self._cache['notype']:
                ntd = self._dao.dNoType(self._pod.dbcon)
                ntd.select_one_deft(msg.notype_id)
                self._cache['notype'][ntd.rec.id] = ntd.rec

                ntempld = self._dao.dNoTempl(self._pod.dbcon)
                ntempld.select_one_deft(msg.notype_id, msg.corrtype_id)

                self._cache['templ'][ntd.rec.id] = self._validate_template(json.loads(ntempld.rec.templ))
                self._cache['res'][ntd.rec.id]   = self._load_template_resources(ntempld.rec.res_list)

            if self._cache['keypair'] is None:
                kpqry = self._dao.dNoDictSearch(self._pod.dbcon)
                kp    = {}
                kpqry.exec()

                while kpqry.fetch():
                    kp[kpqry.orec.id] = kpqry.orec.value or ''

                self._cache['keypair'] = kp

            if self._cache['allow'] is None:
                aqry  = self._dao.dCorrTypeAllowRegexForType(self._pod.dbcon)
                allow = []

                aqry.exec_deft(msg.corrtype_id, '', 'A')

                while aqry.fetch():
                    try:
                        allow.append(re.compile(aqry.orec.regex))
                    except Exception as x:
                        self._log.exception(x)
                        self._log.error('Failed to compile the allow regex: [%s]' % (str(aqry.orec)))
                        raise

                self._cache['allow'] = allow

        except Exception as x:
            self._log.error('Exception caught initializing cache for message: [%s]' % (str(msg)))
            self._log.exception(x)
            return False, str(x)

        return True, None


    def _push_msg_log(self, msglog: str, err_msg: str) -> str:
        """
        Pushes a message onto the message log.
        """
        if not msglog:
            return json.dumps([err_msg])

        try:
            logs = json.loads(msglog)

            logs.insert(0, err_msg)

            if len(logs) > 10:
                logs.pop()

            return json.dumps(logs)

        except Exception as x:
            return json.dumps([str(x)])


    def _load_msgreq_resources(self, req: tMsgReq) -> list:
        """
        Loads all the resources associated with the message.

        :param req: The message req.
        :return: The list of resources loaded or none.
        """
        if not req.cust_res:
            return None

        res = []
        crit = " and mr.corr_id_list ? '%s'" % self.corrtype_id()

        self._dao.dMsgReqResByMsgReq(self._pod.dbcon).exec_deft(req.id, crit).fetch_all(res)

        return res


    def _validate_template(self, templ: dict) -> dict:
        """
        Virtual methed that validates a template for this provider and make any changes as required.

        :param templ: The template configuration.
        :return: The validated template.
        """
        return templ


    def _load_template_resources(self, res_list: str):
        """
        Loads the template resources for by name.

        :param res_list: A json string of resource items.
        :return: The resource objects.
        """
        if not res_list:
            return None

        rlist = json.loads(res_list)

        if not rlist:
            return None

        res = []

        self._dao.dNoResSearch(self._pod.dbcon).exec_deft(QueryBuilder.dynList(rlist, 'nr', 'id')).fetch_all(res)

        return res


    def _build_payload(self, req: tMsgReq) -> dict:
        """
        Pure virtual method creates the message payload for sending.

        :param req: The message request record.
        :return: The payload object.
        """
        raise Exception('_provider_send() not implemented.')


    def _provider_send(self, payload: dict, res_list: list, msg: tMsg, addr_list: list) -> tuple:
        """
        Pure virtual method that sends the pay loda via the correspondence.

        :param payload: The generated payload.
        :param res_list: The list of additional resources to attatch to the message.
        :param msg: The message record.
        :param addr_list: The list of addresses to send to.
        :return: (bool successfailure, float time_taken, str err_msg)
        """
        raise Exception('_provider_send() not implemented.')


    def _template_subst(self, text: str, keypair: dict) -> str:
        """
        Performs standard string subsitution on the payload.

        :param text: The template text to substitute.
        :param keypair: The keypair values to substitute with.
        :return: The substituted string.
        """
        return LocoTemplate(text).substitute(keypair)


    def _build_msg_keypairs(self, req: tMsgReq) -> dict:
        """
        Builds the messasge specific keypairs with the template key pairs combined.

        :param req: The message req object.
        :return: The the msg key pairs.
        """
        res = {}

        templ_kpair = self._cache.get('keypair')

        if templ_kpair:
            res.update(templ_kpair)

        if req.msg_dict:
            kpair = json.loads(req.msg_dict)
            if not isinstance(kpair, dict):
                raise xLoco('Invalid keypair object [msgreq: %s]' % (str(req)))

            res.update(kpair)

        return res


    def _addresses_allowed(self, addr_list: list) -> list:
        """
        Ensure all addresses that are pass the allow list are only returned.

        :param addr_list: The list of addresses we need to check.
        :return: All the addresses that match the regex white list.
        """
        if not self._cache['allow']:
            return addr_list

        res = []

        for chk in addr_list:
            for regexp in self._cache['allow']:
                if regexp.match(chk):
                    res.append(chk)
                    break

        return res



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.config import tConfig

class dConfigInsert:

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tConfig()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    async def exec_deft(self,
                        id: str,
                        descr: str,
                        value: str,
                        grp: str,
                        modified_by: str,
                        tm_stamp: datetime.datetime) -> "dConfigInsert":
        """
        Execute the query by setting all the inputs.

        :param id: str
        :param descr: str
        :param value: str
        :param grp: str
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        :return: Self for convenience.
        """
        self.irec.id = id
        self.irec.descr = descr
        self.irec.value = value
        self.irec.grp = grp
        self.irec.modified_by = modified_by
        self.irec.tm_stamp = tm_stamp

        return await self.exec()

    async def exec(self, irec: tConfig = None) -> "dConfigInsert":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = await self._dbcon.statement("ConfigInsert", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""insert into loco.Config (
  id,
  descr,
  value,
  grp,
  modified_by,
  tm_stamp
) values (
  :id,
  :descr,
  :value,
  :grp,
  :modified_by,
  :tm_stamp
)""")

        self.irec.tm_stamp = await self._dbcon.get_timestamp()

        self._dbstmnt.bind_in("id", self.irec.id, str)
        self._dbstmnt.bind_in("descr", self.irec.descr, str)
        self._dbstmnt.bind_in("value", self.irec.value, str)
        self._dbstmnt.bind_in("grp", self.irec.grp, str)
        self._dbstmnt.bind_in("modified_by", self.irec.modified_by, str)
        self._dbstmnt.bind_in("tm_stamp", self.irec.tm_stamp, datetime.datetime)

        await self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import asyncio
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.config import tConfig
from bs_loco.db.tables.config_key import tConfigKey

from .config_lock_one import dConfigLockOne
from .config_select_one import dConfigSelectOne
from .config_update import dConfigUpdate
from .config_insert import dConfigInsert

class dConfig:

    def __init__(self, dbcon: mettle.db.IAConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self.rec    = tConfig()
        self._dbcon = dbcon

    def __enter__(self):
        """
        DAO enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        DAO exit.
        """
        pass

    async def lock_one(self, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Lock and select the record.

        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        _key = tConfigKey()

        self.rec._write_key(_key)

        return await self.lock_one_by_key(_key, _lock4update, _expected)

    async def lock_one_with(self, _rec: tConfig, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Lock and select the record using a record.

        :param _rec: The table rec to select with and into.
        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        _key = tConfigKey()
        _rec._write_key(_key)

        await self.lock_one_by_key(_key, _lock4update, _expected)
        _rec._copy_from(self.rec)
        return True

    async def lock_one_deft(self,
                            id: str,
                            _lock4update: mettle.db.DBLock,
                            _expected: bool = True) -> bool:
        """
        Attempt to lock and select the record by it's primary key columns.

        :param id: str
        :param _lock4update:  The db lock object to use. 
        :param _expected: bool
        """
        _key = tConfigKey()

        _key.id = id

        return await self.lock_one_by_key(_key, _lock4update, _expected)

    async def lock_one_by_key(self, _key: tConfigKey, _lock4update: mettle.db.DBLock, _expected: bool = True) -> bool:
        """
        Attempt to lock and select the record with it's primary key.

        :param _key: Primary key of the table record.
        :param _lock4update: The db lock object to use.
        :param _expected: If set to false, will not raise if record is not found.
        :return: True if selected, False if not found and _expected is set to False
        """
        with dConfigLockOne(self._dbcon) as _qry:
            _dex = _lock4update.retrys - 1 if _lock4update.retrys > 0 else 0

            while _dex >= 0:
                try:
                    await _qry.exec(_key)
                    if not await _qry.fetch():
                        if not _expected:
                            return False

                        raise mettle.lib.xMettle("LockOne failed to return a record (dConfig)")

                    break
                except mettle.lib.xMettle as _x:
                    if _x.get_error_code() != mettle.lib.xMettle.eCode.DBLockNoWaitFailed:
                        raise _x

                    if _dex == 0:
                        raise mettle.lib.xMettle("Could not lock Config record after %d attempt(s) at %d mili second intervals!" % (
_lock4update.retrys, _lock4update.mili_seconds), errCode = mettle.lib.xMettle.eCode.DBLockNoWaitFailed)

                    if _lock4update.mili_seconds > 0:
                        await asyncio.sleep(_lock4update.mili_seconds * 0.001)
                _dex -= 1

            self.rec = _qry.orec

            return True

    async def try_select_one(self) -> bool:
        """
        Attempt to select the record.

        :return: True if the record was found.
        """
        _key = tConfigKey()

        self.rec._write_key(_key)

        return await self.try_select_one_by_key(_key)

    async def try_select_one_with(self, _rec: tConfig) -> bool:
        """
        Attempt to select the record using a table rec.

        :param _rec: The table rec to select with and into.
        :return: True if the row as selectd.
        """
        _key = tConfigKey()
        _rec._write_key(_key)

        if not await self.try_select_one_by_key(_key):
            return False

        _rec._copy_from(self.rec)

        return True

    async def try_select_one_deft(self,
                                  id: str) -> bool:
        """
        Attempt to select the record by it's primary key columns.

        :param id: str
        :return: True if the row as selectd.
        """
        _key = tConfigKey()

        _key.id = id

        return await self.try_select_one_by_key(_key)

    async def try_select_one_by_key(self, _key: tConfigKey) -> bool:
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :returns: True if row was selected.
        """
        with dConfigSelectOne(self._dbcon) as _qry:
            await _qry.exec(_key)
            if not await _qry.fetch():
                return False

            self.rec = _qry.orec

        return True

    async def select_one(self) -> "dConfig":
        """
        Selects the record, raises and error if it does not exist.

        :return: Self for convenience.
        """
        if not await self.try_select_one():
            raise mettle.lib.xMettle("SelectOne failed to return a record (dConfig)")

        return self

    async def select_one_with(self, _rec: tConfig) -> "dConfig":
        """
        Attempt to select the record with the table record.

        :param _rec: Table record.
        :returns: Self for convenience.
        """
        if not await self.try_select_one_with(_rec):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dConfig)")

        return self

    async def select_one_deft(self,
                              id: str) -> "dConfig":
        """
        Selects the record by its primary key columns, raises and error if it does not exist.

        :param id: str
        """
        _key = tConfigKey()

        _key.id = id

        if not await self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dConfig)")

        return self

    async def select_one_by_key(self, _key: tConfigKey) -> "dConfig":
        """
        Attempt to select the record with it's primary key.

        :param _key: Primary key of the table record.
        :return: Self for convenience.
        """
        if not await self.try_select_one_by_key(_key):
            raise mettle.lib.xMettle("SelectOne failed to return a record (dConfig)")

        return self

    async def update_deft(self,
                          id: str,
                          descr: str,
                          value: str,
                          grp: str,
                          modified_by: str):
        """
        Update the record by setting all column values.

        :param id: str
        :param descr: str
        :param value: str
        :param grp: str
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        """
        self.rec.id = id
        self.rec.descr = descr
        self.rec.value = value
        self.rec.grp = grp
        self.rec.modified_by = modified_by

        await self.update()

    async def update(self, irec: tConfig = None):
        """
        Update the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to update the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dConfigUpdate(self._dbcon) as _qry:
            await _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)

    async def insert_deft(self,
                          id: str,
                          descr: str,
                          value: str,
                          grp: str,
                          modified_by: str):
        """
        Insert the record setting all column values.

        :param id: str
        :param descr: str
        :param value: str
        :param grp: str
        :param modified_by: str
        :param tm_stamp: datetime.datetime
        """
        self.rec.id = id
        self.rec.descr = descr
        self.rec.value = value
        self.rec.grp = grp
        self.rec.modified_by = modified_by

        await self.insert()

    async def insert(self, irec: tConfig = None):
        """
        Insert the record in the database with the record or irec if it is not None.

        :param irec: If not None use this record to insert into the database.
        """
        if irec:
            self.rec._copy_from(irec)

        with dConfigInsert(self._dbcon) as _qry:
            await _qry.exec(self.rec)

        if irec:
            irec._copy_from(self.rec)




# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #


def mimetype_from_content(content: bytes) -> str:
    """
    Gets the mimetype from file content.

    :param content: The content to check.
    :return: The mime type.
    """
    import filetype

    return filetype.guess(content)


def mimetype_from_filename(filename: str) -> str:
    """
    Gets the mimetype from a file name.

    :param content: The content to check.
    :return: The mime type.
    """
    import mimetypes

    mtype, encoding = mimetypes.guess_type(filename)

    return mtype



# **************************************************************************** #
#                           This file is part of:                              #
#                                BITSMITHS                                     #
#                           https://bitsmiths.co.za                            #
# **************************************************************************** #
#  Copyright (C) 2015 - 2021 Bitsmiths (Pty) Ltd.  All rights reserved.        #
#   * https://bitbucket.org/bitsmiths_za/bitsmiths                             #
#                                                                              #
#  Permission is hereby granted, free of charge, to any person obtaining a     #
#  copy of this software and associated documentation files (the "Software"),  #
#  to deal in the Software without restriction, including without limitation   #
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,    #
#  and/or sell copies of the Software, and to permit persons to whom the       #
#  Software is furnished to do so, subject to the following conditions:        #
#                                                                              #
#  The above copyright notice and this permission notice shall be included in  #
#  all copies or substantial portions of the Software.                         #
#                                                                              #
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR  #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,    #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL     #
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER  #
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING     #
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER         #
#  DEALINGS IN THE SOFTWARE.                                                   #
# **************************************************************************** #
import datetime
import json

from bs_lib import Pod

import bs_loco

from bs_loco       import util
from bs_loco.xloco import xLoco

from bs_loco.db.tables    import tMsg
from bs_loco.db.tables    import tMsgReq

from .loco_trig import LocoTrig


class LocalDB(LocoTrig):
    """
    The local database trigger implementation.
    """

    def __init__(self):
        """
        Constructor.
        """
        LocoTrig.__init__(self)
        self._dao = None


    def id(self) -> str:
        """
        Overload.
        """
        raise 'local_db'


    def name(self) -> str:
        """
        Overload.
        """
        'Local DB'


    def initialize(self, pod: Pod, cfg: dict):
        """
        Overload.
        """
        LocoTrig.initialize(self, pod, cfg)

        if not cfg or not cfg.get('dao'):
            raise xLoco('Local DB requires a dao config value.')

        self._dao = bs_loco.dao_by_name(cfg.get('dao'))


    def destroy(self):
        """
        Overload.
        """
        LocoTrig.destroy(self)
        self._dao = None


    def _trigger(self, notype: str, corr_addr: dict, key_pair: dict = None, res: dict = None, meta_data: dict = None) -> int:
        """
        Overload.
        """
        mreq     = self._dao.dMsgReq(self._pod.dbcon)
        mres     = None
        nt       = self._dao.dNoType(self._pod.dbcon)
        dtnow    = datetime.datetime.now()
        corrlist = []
        cnt      = 0

        nt.select_one_deft(notype)

        if nt.rec.pref_corr:
            corrlist = json.loads(nt.rec.pref_corr)

            for corr in corr_addr:
                if corr not in corrlist:
                    corrlist.append(corr)

        else:
            corrlist = list(corr_addr.keys())

        for corr in corrlist:
            if nt.rec.corr_cnt and cnt >= nt.rec.corr_cnt:
                self._pod.log.info('Max correspondece count reached [%d], skipped: %s' % (
                    nt.rec.corr_cnt, str(corrlist[cnt:])))
                break

            cnt += 1

        mreq.rec.clear()

        mreq.insert_deft(notype,
                         nt.rec.priority,
                         tMsgReq.Status_Couplet.key_pending,
                         dtnow,
                         dtnow,
                         datetime.datetime.min,
                         json.dumps(corr_addr),
                         '' if not key_pair else json.dumps(key_pair),
                         True if res and res.get('resources') else False,
                         '' if not meta_data else json.dumps(meta_data),
                         self._pod.usr_id or '[Anonymous]')

        if res and res.get('resources'):
            for ritem in res['resources']:

                fnd = False

                for fc in ritem['for']:
                    if fc in corr:
                        fnd = True
                        break

                if not fnd:
                    continue

                if not mres:
                    mres = self._dao.dMsgReqRes(self._pod.dbcon)

                mres.rec.clear()

                if isinstance(ritem['content'], str):
                    content = bytes(ritem['content'], 'utf8')
                else:
                    content = ritem['content']

                mres.insert_deft(mreq.rec.id,
                                 json.dumps(ritem['for']),
                                 ritem['filename'],
                                 ritem.get('mime') or util.mimetype_from_content(content),
                                 content,
                                 self._pod.usr_id)


        msg = self._dao.dMsg(self._pod.dbcon)

        for corr in corrlist:
            msg.insert_deft(mreq.rec.id,
                            corr,
                            notype,
                            mreq.rec.priority,
                            tMsg.Status_Couplet.key_pending,
                            0,
                            0,
                            datetime.datetime.min,
                            0.0,
                            '',
                            '',
                            '',
                            self._pod.usr_id)

        return len(corrlist)



# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#

import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

class tMsgKey(mettle.io.ISerializable):

    def __init__(self,
                 id: int = 0):
        """
        Constructor.

        :param id: int
        """
        self.id = id

    def clear(self):
        """
        Clears all member variables for this database record.
        """
        self.id = 0

    def _name(self) -> str:
        """
        Name of the record.

        :return: The name.
        """
        return 'tMsgKey'

    def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
        """
        Serialize record to a stream.

        :param _w:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _w.write_start(_oname)
        _w.write_int64("id", self.id)
        _w.write_end(_oname)

    def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
        """
        Deserialize record from a stream.

        :param _r:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _r.read_start(_oname)
        self.id = _r.read_int64("id")
        _r.read_end(_oname)

    def _copy_from(self, rec: "tMsgKey"):
        """
        Copies the column from the rec into this record.

        :param rec: The source record.
        """
        self.id = rec.id

    @staticmethod
    def _cache_davs(dvc=None):
        """
        Cache the DAV into the dav cache (or create one) and return it.

        :param dvc: (mettle.lib.DavCache), target cache, if None a new one is created and returned.
        """
        if dvc is None:
            dvc = mettle.lib.DavCache()

        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.Min, 1))
        dvc.add_targ("id", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))

        return dvc

    def _get_davs(self, dvc=None):
        tMsgKey._cache_davs(dvc)

    @staticmethod
    def _pk_compare(x, y) -> int:
        """
        Compare function do primary key comparisons for list sorts and searches.

        :return: zero for x == y, negative for x < y, positive for x > y
        """
        return x.id - y.id

    def __lt__(self, other):
        return other and self._pk_compare(self, other) < 0

    def __gt__(self, other):
        return other and self._pk_compare(self, other) > 0

    def __eq__(self, other):
        return other and self._pk_compare(self, other) == 0

    def __le__(self, other):
        return other and self._pk_compare(self, other) <= 0

    def __ge__(self, other):
        return other and self._pk_compare(self, other) >= 0

    def __ne__(self, other):
        return other == None or self._pk_compare(self, other)

    def __repr__(self) -> str:
        return '<Msg [id:%r]>' % (
            self.id)


    class List(list, mettle.io.ISerializable):
        """
        List Class
        """

        def _name(self) -> str:
            """
            Name of the list.

            :return: The name
            """
            return 'tMsgKey.List'

        def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
            """
            Serialize the list to a stream.

            :param _w:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _w.write_start_list(_oname, len(self))

            for _rec in self:
                _rec._serialize(_w)

            _w.write_end(_oname)

        def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
            """
            Deserialize the list from a stream.

            :param _r:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _cnt = _r.read_start_list(_oname)

            while _cnt >= 1:
                _rec  = tMsgKey()
                _cnt -= 1
                _rec._deserialize(_r)
                self.append(_rec)

            _r.read_end(_oname)

        def __repr__(self) -> str:
            return '<tMsgKey.List[%d]>' % (len(self))

        def _get_davs(self, dvc=None):
            return tMsgKey._cache_davs(dvc)




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#  Target Database = postgresql
#
import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from bs_loco.db.tables.msg_key import tMsgKey

class dMsgDeleteOne:

    def __init__(self, dbcon: mettle.db.IConnect):
        """
        Constructor.

        :param dbcon: Mettle database connection object.
        """
        self._dbcon   = dbcon
        self._dbstmnt = None
        self.irec     = tMsgKey()

    def __del__(self):
        """
        Destructor.
        """
        del self._dbstmnt

    def __enter__(self):
        """
        With statement enter.
        """
        return self

    def __exit__(self, type, value, traceback):
        """
        With statement exit.
        """
        self._destroy()

    def _destroy(self):
        if self._dbstmnt == None:
            return

        del self._dbstmnt
        self._dbstmnt = None

    def exec_deft(self,
                  id: int) -> "dMsgDeleteOne":
        """
        Execute the query by setting all the inputs.

        :param id: int
        :return: Self for convenience.
        """
        self.irec.id = id

        return self.exec()

    def exec(self, irec: tMsgKey = None) -> "dMsgDeleteOne":
        """
        Execute the query, optionally passing in the input rec.

        :param irec:
        :return: Self for convenience.
        """
        if irec:
            self.irec._copy_from(irec)

        self._destroy()

        self._dbstmnt = self._dbcon.statement("MsgDeleteOne", self._dbcon.STMNT_TYPE_CUD)

        self._dbstmnt.sql("""delete from
  loco.Msg
where
  id = :id""")

        self._dbstmnt.bind_in("id", self.irec.id, int, 8)

        self._dbcon.execute(self._dbstmnt)

        if irec:
            irec._copy_from(self.irec)

        return self




# This file was generated by mettle.genes.db.GenPy3 [ver 2.1] on Mon Feb 15 23:21:02 2021
#

import copy
import datetime
import uuid
import time
import mettle.lib
import mettle.io
import mettle.db

from .snap_msg_req_key import tSnapMsgReqKey

class tSnapMsgReq(mettle.io.ISerializable, mettle.db.IPrimaryKey):

    def __init__(self,
                 valuedate: datetime.date = datetime.date.min,
                 notype_id: str = '',
                 tot_req: int = 0,
                 tot_proc: int = 0,
                 tot_cancelled: int = 0,
                 tot_part_canc: int = 0,
                 tot_msgs: int = 0,
                 tot_msgs_sent: int = 0,
                 tot_msgs_failed: int = 0,
                 tot_msgs_retrys: int = 0):
        """
        Constructor.

        :param valuedate: datetime.date
        :param notype_id: str
        :param tot_req: int
        :param tot_proc: int
        :param tot_cancelled: int
        :param tot_part_canc: int
        :param tot_msgs: int
        :param tot_msgs_sent: int
        :param tot_msgs_failed: int
        :param tot_msgs_retrys: int
        """
        self.valuedate = valuedate
        self.notype_id = notype_id
        self.tot_req = tot_req
        self.tot_proc = tot_proc
        self.tot_cancelled = tot_cancelled
        self.tot_part_canc = tot_part_canc
        self.tot_msgs = tot_msgs
        self.tot_msgs_sent = tot_msgs_sent
        self.tot_msgs_failed = tot_msgs_failed
        self.tot_msgs_retrys = tot_msgs_retrys

    def clear(self):
        """
        Clears all member variables for this database record.
        """
        self.valuedate = datetime.date.min
        self.notype_id = ''
        self.tot_req = 0
        self.tot_proc = 0
        self.tot_cancelled = 0
        self.tot_part_canc = 0
        self.tot_msgs = 0
        self.tot_msgs_sent = 0
        self.tot_msgs_failed = 0
        self.tot_msgs_retrys = 0

    def _name(self) -> str:
        """
        Name of the record.

        :return: The name.
        """
        return 'tSnapMsgReq'

    def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
        """
        Serialize record to a stream.

        :param _w:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _w.write_start(_oname)
        _w.write_date("valuedate", self.valuedate)
        _w.write_string("notype_id", self.notype_id)
        _w.write_int32("tot_req", self.tot_req)
        _w.write_int32("tot_proc", self.tot_proc)
        _w.write_int32("tot_cancelled", self.tot_cancelled)
        _w.write_int32("tot_part_canc", self.tot_part_canc)
        _w.write_int32("tot_msgs", self.tot_msgs)
        _w.write_int32("tot_msgs_sent", self.tot_msgs_sent)
        _w.write_int32("tot_msgs_failed", self.tot_msgs_failed)
        _w.write_int32("tot_msgs_retrys", self.tot_msgs_retrys)
        _w.write_end(_oname)

    def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
        """
        Deserialize record from a stream.

        :param _r:
        :param _oname:
        """
        if _oname == None:
            _oname = self._name()

        _r.read_start(_oname)
        self.valuedate = _r.read_date("valuedate")
        self.notype_id = _r.read_string("notype_id")
        self.tot_req = _r.read_int32("tot_req")
        self.tot_proc = _r.read_int32("tot_proc")
        self.tot_cancelled = _r.read_int32("tot_cancelled")
        self.tot_part_canc = _r.read_int32("tot_part_canc")
        self.tot_msgs = _r.read_int32("tot_msgs")
        self.tot_msgs_sent = _r.read_int32("tot_msgs_sent")
        self.tot_msgs_failed = _r.read_int32("tot_msgs_failed")
        self.tot_msgs_retrys = _r.read_int32("tot_msgs_retrys")
        _r.read_end(_oname)

    def _copy_from(self, rec: "tSnapMsgReq"):
        """
        Copies the column from the rec into this record.

        :param rec: The source record.
        """
        self.valuedate = rec.valuedate
        self.notype_id = rec.notype_id
        self.tot_req = rec.tot_req
        self.tot_proc = rec.tot_proc
        self.tot_cancelled = rec.tot_cancelled
        self.tot_part_canc = rec.tot_part_canc
        self.tot_msgs = rec.tot_msgs
        self.tot_msgs_sent = rec.tot_msgs_sent
        self.tot_msgs_failed = rec.tot_msgs_failed
        self.tot_msgs_retrys = rec.tot_msgs_retrys

    def _write_pk(self, _key: "tSnapMsgReq"):
        """
        Write the primary key columns to the key object.

        :param _key:
        """
        self._write_key(_key)

    def _read_pk(self, _key: "tSnapMsgReq"):
        """
        Read the primary key columns from the key object.

        :param _key:
        """
        self._read_key(_key)

    def _serialize_key(self, _w: mettle.io.IWriter):
        """
        Serialize the primary key to a stream.

        :param _w:
        """
        _key = tSnapMsgReqKey()

        self._write_key(_key)
        _key._serialize(_w)

    def _deserialize_key(self, _r: mettle.io.IReader):
        """
        Deserialize the primary key from a stream.

        :param _r:
        """
        _key = tSnapMsgReqKey()

        _key._deserialize(_r)
        self._read_key(_key)

    @staticmethod
    def _cache_davs(dvc=None, for_ins: bool = False, key_only: bool = False):
        """
        Cache the DAV into the dav cache (or create one) and return it.

        :param dvc: (mettle.lib.DavCache), target cache, if None a new one is created and returned.
        :param for_ins: Davs for insert.
        :param key_only: Davs by key only, ie for delete/selectone.
        """
        if dvc is None:
            dvc = mettle.lib.DavCache()

        dvc.add_targ("valuedate", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("notype_id", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("notype_id", mettle.lib.Dav(mettle.lib.Dav.eDavType.Max, 64))
        if key_only:
            return dvc

        dvc.add_targ("tot_req", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("tot_proc", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("tot_cancelled", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("tot_part_canc", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("tot_msgs", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("tot_msgs_sent", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("tot_msgs_failed", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))
        dvc.add_targ("tot_msgs_retrys", mettle.lib.Dav(mettle.lib.Dav.eDavType.NotNull))

        return dvc

    def _get_davs(self, dvc=None, for_ins: bool = False, key_only: bool = False):
        tSnapMsgReq._cache_davs(dvc, for_ins, key_only)

    def _write_key(self, _key: tSnapMsgReqKey):
        """
        Write primary key columns to _key.

        :param _key:
        """
        _key.valuedate = self.valuedate
        _key.notype_id = self.notype_id

    def _read_key(self, _key: tSnapMsgReqKey):
        """
        Read the primary key columns from _key.

        :param _key:
        """
        self.valuedate = _key.valuedate
        self.notype_id = _key.notype_id

    @staticmethod
    def _pk_compare(x, y) -> int:
        """
        Compare function do primary key comparisons for list sorts and searches.

        :return: zero for x == y, negative for x < y, positive for x > y
        """
        rc = (x.valuedate > y.valuedate) - (x.valuedate < y.valuedate)

        if rc != 0:
            return rc

        rc = (x.notype_id > y.notype_id) - (x.notype_id < y.notype_id)

        return rc

    def __lt__(self, other):
        return other and self._pk_compare(self, other) < 0

    def __gt__(self, other):
        return other and self._pk_compare(self, other) > 0

    def __eq__(self, other):
        return other and self._pk_compare(self, other) == 0

    def __le__(self, other):
        return other and self._pk_compare(self, other) <= 0

    def __ge__(self, other):
        return other and self._pk_compare(self, other) >= 0

    def __ne__(self, other):
        return other == None or self._pk_compare(self, other)

    def __repr__(self) -> str:
        return '<SnapMsgReq [valuedate:%s, notype_id:%.32s, tot_req:%r, tot_proc:%r, tot_cancelled:%r, tot_part_canc:%r, tot_msgs:%r, tot_msgs_sent:%r, tot_msgs_failed:%r, tot_msgs_retrys:%r]>' % (
            self.valuedate, self.notype_id, self.tot_req, self.tot_proc, self.tot_cancelled, self.tot_part_canc, self.tot_msgs, self.tot_msgs_sent, self.tot_msgs_failed, self.tot_msgs_retrys)


    class List(list, mettle.io.ISerializable):
        """
        List Class
        """

        def _name(self) -> str:
            """
            Name of the list.

            :return: The name
            """
            return 'tSnapMsgReq.List'

        def _serialize(self, _w: mettle.io.IWriter, _oname: str = None):
            """
            Serialize the list to a stream.

            :param _w:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _w.write_start_list(_oname, len(self))

            for _rec in self:
                _rec._serialize(_w)

            _w.write_end(_oname)

        def _deserialize(self, _r: mettle.io.IReader, _oname: str = None):
            """
            Deserialize the list from a stream.

            :param _r:
            :param _oname:
            """
            if _oname == None:
                _oname = self._name()

            _cnt = _r.read_start_list(_oname)

            while _cnt >= 1:
                _rec  = tSnapMsgReq()
                _cnt -= 1
                _rec._deserialize(_r)
                self.append(_rec)

            _r.read_end(_oname)

        def __repr__(self) -> str:
            return '<tSnapMsgReq.List[%d]>' % (len(self))

        def _get_davs(self, dvc=None, for_ins: bool = False, key_only: bool = False):
            return tSnapMsgReq._cache_davs(dvc, for_ins, key_only)

